import asyncio
from enum import Enum


class Gate:
    def __init__(self, n=32, *, loop=None):
        loop = loop if loop else asyncio.get_event_loop()
        self._loop = loop
        self._res = asyncio.Semaphore(n, loop=loop)
        self._state = asyncio.Event(loop=loop)
        self._emtpy_statue = asyncio.Event(loop=loop)
        self._emtpy_statue.set()
        self._counter = 0

    async def acquire(self):
        await self._state.wait()
        self._emtpy_statue.clear()
        self._counter += 1
        await self._res.acquire()
        return True

    async def release(self):
        self._res.release()
        self._counter -= 1
        if self._counter == 0:
            self._emtpy_statue.set()

    async def __aenter__(self):
        await self.acquire()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.release()

    async def open(self):
        self._state.set()

    async def close(self):
        self._state.clear()

    async def close_until_complete(self):
        self._state.clear()
        await self._emtpy_statue.wait()


class ServerStatusEnum(Enum):
    Zero = 0
    Running = 1
    NotAvailable = 2
    Updating = 3


class ServerStatus:
    def __init__(self, gate: Gate = None):
        self._gate = gate or Gate()
        self._state = ServerStatusEnum.Zero
        self.gate = self._gate

    async def error(self):
        self._state = ServerStatusEnum.NotAvailable
        await self.gate.close_until_complete()

    async def ready(self):
        self._state = ServerStatusEnum.Running
        await self.gate.open()

    async def left(self):
        self._state = ServerStatusEnum.Zero
        await self.gate.close_until_complete()

    async def update(self):
        self._state = ServerStatusEnum.Updating
        await self.gate.close_until_complete()


__all__ = ["Gate", "ServerStatusEnum", "ServerStatus", ]



from typing import *
import asyncio
from .error import *
from .result import Result


class RedisTaskNode:
    def __init__(self, command, fa: Optional[Tuple[Callable, Union[None, Dict]]], writer_queue: asyncio.Queue):
        self.command: bytes = command
        self.fa = fa
        self.writer_queue = writer_queue
        self.result = None

    @classmethod
    async def create(cls, reader, writer_queue: asyncio.Queue, need_auth_event: asyncio.Event, mixin: 'RedisServerBase') -> Result['RedisTaskNode']:
        idle_timeout = mixin.IDLE_TIMEOUT
        line = await asyncio.wait_for(reader.readline(), timeout=idle_timeout)
        if not line:
            return Result(EOFError())
        args_count_result = mixin.get_count(line)
        if args_count_result.is_error:
            return Result(args_count_result.error)
        args_count = args_count_result.unwrap()
        commands = list()
        for _ in range(args_count):
            data = await asyncio.wait_for(reader.readline(), timeout=idle_timeout)
            binary_size_result = mixin.get_binary_size(data)
            if binary_size_result.is_error:
                return Result(binary_size_result.error)
            binary_size = binary_size_result.unwrap()
            data = await asyncio.wait_for(reader.readexactly(binary_size + 2), timeout=idle_timeout)
            commands.append(data[:-2])
        if not commands:
            return Result(RedisProtocolFormatError())
        command = commands[0].upper()
        if command == b'AUTH':
            if len(commands) > 1:
                password = commands[1]
            else:
                password = b''
            fa = (mixin.auth, dict(password=password), )
            node = RedisTaskNode(command, fa, writer_queue)
            return Result(node)
        if need_auth_event.is_set():
            node = RedisTaskNode(b'AUTH', None, writer_queue)
            node.result = Result(NoPasswordError("NOAUTH Authentication required."))
            return Result(node)
        if command == b'CONFIG':
            if len(commands) != 3:
                return Result(RedisProtocolFormatError())
            fa = (mixin.config, dict(get_set=commands[1], field=commands[2]))
            node = RedisTaskNode(command, fa, writer_queue)
            return Result(node)
        elif command == b'INFO':
            fa = (mixin.info, None)
            node = RedisTaskNode(command, fa, writer_queue)
            return Result(node)
        elif command == b'SELECT':
            if len(commands) != 2:
                return Result(RedisProtocolFormatError())
            database_id = int(commands[1])
            fa = (mixin.select, dict(db=database_id))
            node = RedisTaskNode(command, fa, writer_queue)
            return Result(node)
        elif command == b'DBSIZE':
            if len(commands) != 1:
                return Result(RedisProtocolFormatError())
            fa = (mixin.dbsize, None)
            node = RedisTaskNode(command, fa, writer_queue)
            return Result(node)
        elif command == b'PING':
            fa = (mixin.ping, None)
            node = RedisTaskNode(command, fa, writer_queue)
            return Result(node)
        elif command == b'SET':
            if len(commands) < 3:
                return Result(RedisProtocolFormatError())
            key = commands[1].decode("utf8").strip()
            value = commands[2]
            fa = (mixin.set, dict(key=key, value=value))
            node = RedisTaskNode(command, fa, writer_queue)
            return Result(node)
        elif command == b'GET':
            if len(commands) != 2:
                return Result(RedisProtocolFormatError())
            key = commands[1].decode("utf8").strip()
            fa = (mixin.get, dict(key=key))
            node = RedisTaskNode(command, fa, writer_queue)
            return Result(node)
        elif command == b'TYPE':
            if len(commands) != 2:
                return Result(RedisProtocolFormatError())
            key = commands[1].decode("utf8").strip()
            fa = (mixin.key_type, dict(key=key))
            node = RedisTaskNode(command, fa, writer_queue)
            return Result(node)
        elif command == b'TTL':
            if len(commands) != 2:
                return Result(RedisProtocolFormatError())
            key = commands[1].decode("utf8").strip()
            fa = (mixin.ttl, dict(key=key))
            node = RedisTaskNode(command, fa, writer_queue)
            return Result(node)
        elif command == b'SCAN':
            if len(commands) < 2 or len(commands) % 2:
                return Result(RedisProtocolFormatError())
            pattern = None
            args_pairs = zip(commands[2::2], commands[3::2])
            for arg_name, arg_value in args_pairs:
                if arg_name.upper() == b'MATCH':
                    pattern = f'^{arg_value.decode("utf8").replace("*", ".+")}$'
                elif arg_name.upper() == b'COUNT':
                    pass
            fa = (mixin.scan, dict(pattern=pattern))
            node = RedisTaskNode(command, fa, writer_queue)
            return Result(node)
        else:
            node = RedisTaskNode(b'NOOP', None, writer_queue)
            node.result = Result(CommandNotFound())
            return Result(node)

    async def write_result(self, writer, need_auth_event: asyncio.Event, mixin: 'RedisServerMixin') -> Result[bool]:
        command = self.command
        if command == b'AUTH':
            auth_result = self.result
            if auth_result.is_error:
                if isinstance(auth_result.error, ExceptionWithReplyError):
                    error_reply = auth_result.error.reply
                else:
                    error_reply = "ERR"
                binary_result = mixin.set_err(error_reply)
                need_auth_event.set()
            elif not auth_result.unwrap():
                binary_result = mixin.set_err("ERR invalid password")
                need_auth_event.set()
            else:
                binary_result = mixin.set_ok()
                need_auth_event.clear()
            writer.write(binary_result.unwrap())
        elif need_auth_event.is_set():
            binary_result = mixin.set_err("NOAUTH Authentication required.")
            writer.write(binary_result.unwrap())
        elif command == b'CONFIG':
            config_result = self.result
            if config_result.is_error:
                return Result(config_result.error)
            binary_lines = config_result.unwrap()
            item_size = len(binary_lines)
            writer.write(mixin.set_count(item_size).unwrap())
            for binary in binary_lines:
                binary_result = mixin.set_binary(binary)
                writer.write(binary_result.unwrap())
        elif command == b'INFO':
            info_result = self.result
            if info_result.is_error:
                return Result(info_result.error)
            binary_result = mixin.set_binary(info_result.unwrap())
            if binary_result.is_error:
                return Result(binary_result.error)
            writer.write(binary_result.unwrap())
        elif command == b'SELECT':
            select_result = self.result
            if select_result.is_error:
                writer.write(mixin.set_err("ERR invalid DB index").unwrap())
            else:
                writer.write(mixin.set_ok().unwrap())
        elif command == b'DBSIZE':
            dbsize_result = self.result
            if dbsize_result.is_error:
                return Result(dbsize_result.error)
            binary_result = mixin.set_integer(dbsize_result.unwrap())
            if binary_result.is_error:
                return Result(binary_result.error)
            writer.write(binary_result.unwrap())
        elif command == b'PING':
            ping_result = self.result
            if ping_result.is_error:
                return Result(ping_result.error)
            binary_result = mixin.set_ok(ping_result.unwrap())
            if binary_result.is_error:
                return Result(binary_result.error)
            writer.write(binary_result.unwrap())
        elif command == b'SET':
            set_result = self.result
            if set_result.is_error:
                return Result(set_result.error)
            binary_result = mixin.set_ok()
            if binary_result.is_error:
                return Result(binary_result.error)
            writer.write(binary_result.unwrap())
        elif command == b'GET':
            get_result = self.result
            if get_result.is_error:
                return Result(get_result.error)
            if get_result.unwrap() is None:
                binary_result = mixin.set_nil()
            else:
                binary_result = mixin.set_binary(get_result.unwrap())
            if binary_result.is_error:
                return Result(binary_result.error)
            writer.write(binary_result.unwrap())
        elif command == b'TYPE':
            key_type_result = self.result
            if key_type_result.is_error:
                return Result(key_type_result.error)
            binary_result = mixin.set_ok(key_type_result.unwrap())
            if binary_result.is_error:
                return Result(binary_result.error)
            writer.write(binary_result.unwrap())
        elif command == b'TTL':
            ttl_result = self.result
            if ttl_result.is_error:
                return Result(ttl_result.error)
            binary_result = mixin.set_integer(ttl_result.unwrap())
            if binary_result.is_error:
                return Result(binary_result.error)
            writer.write(binary_result.unwrap())
        elif command == b'SCAN':
            scan_result = self.result
            if scan_result.is_error:
                return Result(scan_result.error)
            keys = scan_result.unwrap()
            count = len(keys)
            writer.write(mixin.set_count(2).unwrap())
            writer.write(mixin.set_binary(0).unwrap())
            writer.write(mixin.set_count(count).unwrap())
            for key in keys:
                writer.write(mixin.set_binary(key).unwrap())
                await writer.drain()
        else:
            binary_result = mixin.set_err("COMMAND NOT EXISTS")
            writer.write(binary_result.unwrap())
        await writer.drain()
        return Result(True)


__all__ = ["RedisTaskNode", ]



from typing import *
import asyncio
from .result import *
from .redis_protocol import RedisProtocol
from .redis_task_node import RedisTaskNode


class RedisServerBase(RedisProtocol):
    IDLE_TIMEOUT = 60
    WORKER_QUEUE = asyncio.Queue()
    ENABLE_AUTH = False

    def __init__(self):
        self._http_app = None

    @property
    def app(self):
        return self._http_app

    async def worker(self):
        while True:
            try:
                node: RedisTaskNode = await self.WORKER_QUEUE.get()
                if node.fa:
                    func, kwargs = node.fa
                    if kwargs:
                        result = await func(**kwargs)
                    else:
                        result = await func()
                    node.result = result
                await node.writer_queue.put(node)
            except asyncio.CancelledError:
                return
            except Exception as e:
                pass

    async def handle_stream(self, reader, writer):
        try:
            auth = self.ENABLE_AUTH
            need_auth = True if auth else False
            need_auth_event = asyncio.Event()
            if need_auth:
                need_auth_event.set()
            writer_queue = asyncio.Queue()
            await asyncio.wait([
                self.reader_loop(reader, need_auth_event, writer_queue),
                self.writer_loop(writer, need_auth_event, writer_queue),
            ])
        except asyncio.TimeoutError as e:
            return Result(e)
        except Exception as e:
            return Result(e)
        finally:
            try:
                writer.close()
            except Exception as e:
                pass

    async def reader_loop(self, reader, need_auth_event: asyncio.Event, writer_queue: asyncio.Queue):
        try:
            while True:
                node_result = await RedisTaskNode.create(reader, writer_queue, need_auth_event, self)
                if node_result.is_error:
                    raise node_result.error
                if node_result.is_some:
                    await self.WORKER_QUEUE.put(node_result.unwrap())
        except asyncio.TimeoutError as e:
            return Result(e)
        except Exception as e:
            return Result(e)
        finally:
            await writer_queue.put(None)

    async def writer_loop(self, writer, need_auth_event: asyncio.Event, writer_queue: asyncio.Queue):
        idle_timeout = self.IDLE_TIMEOUT
        try:
            while True:
                node: Optional[RedisTaskNode] = await writer_queue.get()
                if node is None:
                    return
                write_result = await asyncio.wait_for(node.write_result(writer,  need_auth_event, self), timeout=idle_timeout)
                if write_result.is_error:
                    raise write_result.error
                elif not write_result.unwrap():
                    return
        except asyncio.TimeoutError as e:
            return Result(e)
        except Exception as e:
            return Result(e)
        finally:
            try:
                writer.close()
                await writer.wait_closed()
            except Exception as e:
                pass

    async def serve(self, host='127.0.0.1', port=6379):
        server = await asyncio.start_server(self.handle_stream, host, port)
        async with server:
            await asyncio.gather(server.serve_forever(), self.worker())

    async def get(self, key) -> Result[bytes]:
        return Result(NotImplementedError())

    async def getset(self, key, value) -> Result:
        return Result(NotImplementedError())

    async def set(self, key, value) -> Result:
        return Result(NotImplementedError())

    def hmget(self, key, *fields):
        return Result(NotImplementedError())

    def hmset(self, key, **fields):
        return Result(NotImplementedError())

    async def key_type(self, key):
        return Result(NotImplementedError())

    async def ping(self):
        return Result(NotImplementedError())

    async def info(self):
        return Result(NotImplementedError())

    def subscribe(self, channle):
        return Result(NotImplementedError())

    async def auth(self, password) -> Result[bool]:
        """
        如果是连接状态问题或者非密码对错的情况下，这里应返回 Result[Exception]
        判断密码正确时，应返回 Result[bool]
        :param password:
        :return:
        """
        return Result(NotImplementedError())

    async def ttl(self, key):
        return Result(NotImplementedError())

    async def scan(self, pattern):
        return Result(NotImplementedError())

    async def dbsize(self):
        return Result(NotImplementedError())

    async def select(self, db):
        return Result(NotImplementedError())

    async def config(self, get_set, field, value=None):
        return Result(NotImplementedError())


__all__ = ["RedisServerBase", ]



from typing import *
import aiohttp
from aiohttp import web
from .websocket_session import WebsocketSession


class WebsocketApp(web.Application):
    def __init__(self, session_class=WebsocketSession, **kwargs):
        super(WebsocketApp, self).__init__(**kwargs)
        self._session_class = session_class
        self._session_dict: Dict[str, WebsocketSession] = dict()

    def all_sessions(self):
        return list(self._session_dict.values())

    def __getitem__(self, item) -> WebsocketSession:
        return self._session_dict.__getitem__(item)

    def __contains__(self, item):
        return self._session_dict.__contains__(item)

    async def handler(self, request):
        if not self._session_class:
            return
        ws = web.WebSocketResponse()
        await ws.prepare(request)
        session = self._session_class()
        session.socket = ws
        self._session_dict[session.session_id] = session
        await session.prepare()
        try:
            async for msg in ws:
                if msg.type == aiohttp.WSMsgType.TEXT or msg.type == aiohttp.WSMsgType.BINARY:
                    message_result = session.message_loads(msg.data)
                    if message_result.is_some:
                        await session.request(message_result.unwrap())
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    break
        finally:
            try:
                await ws.close()
            except Exception as e:
                pass
            try:
                await session.on_finish()
            except Exception as e:
                pass
            if session.session_id in self._session_dict:
                del self._session_dict[session.session_id]
        return ws


__all__ = ["WebsocketApp", ]





class RedisProtocolFormatError(Exception):
    pass


class ExceptionWithReplyError(Exception):
    def __init__(self, reply="ERR", *args):
        super().__init__(*args)
        self.reply = reply


class NoPasswordError(ExceptionWithReplyError):
    pass


class ReducerTypeNotFound(Exception):
    pass


class StoreNotFound(Exception):
    pass


class NodeNotFound(Exception):
    pass


class ReducerNotFound(Exception):
    pass


class DatabaseNotFound(Exception):
    pass


class CommandNotFound(ExceptionWithReplyError):
    def __init__(self):
        super(CommandNotFound, self).__init__(reply="COMMAND NOT FOUND")


__all__ = [
    "RedisProtocolFormatError",
    "ExceptionWithReplyError",
    "NoPasswordError",
    "ReducerTypeNotFound",
    "StoreNotFound",
    "NodeNotFound",
    "ReducerNotFound",
    "DatabaseNotFound",
    "CommandNotFound",
]



from typing import *
from .result import *
from .redis_server_base import RedisServerBase


class RedisServer(RedisServerBase):
    async def get(self, key) -> Result[bytes]:
        return Result(NotImplementedError())

    async def getset(self, key, value) -> Result:
        return Result(NotImplementedError())

    async def set(self, key, value) -> Result:
        return Result(NotImplementedError())

    def hmget(self, key, *fields):
        return Result(NotImplementedError())

    def hmset(self, key, **fields):
        return Result(NotImplementedError())

    async def key_type(self, key):
        return Result(NotImplementedError())

    async def ping(self):
        return Result(NotImplementedError())

    async def info(self):
        return Result(NotImplementedError())

    def subscribe(self, channle):
        return Result(NotImplementedError())

    async def auth(self, password) -> Result[bool]:
        """
        如果是连接状态问题或者非密码对错的情况下，这里应返回 Result[Exception]
        判断密码正确时，应返回 Result[bool]
        :param password:
        :return:
        """
        return Result(NotImplementedError())

    async def ttl(self, key):
        return Result(NotImplementedError())

    async def scan(self, pattern):
        return Result(NotImplementedError())

    async def dbsize(self):
        return Result(NotImplementedError())

    async def select(self, db):
        return Result(NotImplementedError())

    async def config(self, get_set, field, value=None):
        return Result(NotImplementedError())


__all__ = ["RedisServer", ]



from .result import Result
from .error import *
from .gate import Gate
from .redis_protocol import RedisProtocol
from .redis_server_base import RedisServerBase
from .redis_server import RedisServer
from .websocket_session import WebsocketSession
from .websocket_app import WebsocketApp
from .node import PorkPepperNode



from typing import *


T = TypeVar('T')
E = TypeVar('E')


class Result(Generic[T]):
    def __init__(self, v: Any = None):
        self._v = v

    @property
    def is_ok(self):
        return not isinstance(self._v, Exception)

    @property
    def is_error(self) -> bool:
        return isinstance(self._v, Exception)

    @property
    def is_none(self) -> bool:
        return self._v is None

    @property
    def is_some(self) -> bool:
        return not self.is_none and not self.is_error

    def unwrap(self) -> T:
        if self.is_error:
            raise ReferenceError
        return self._v

    @property
    def error(self) -> Optional[Exception]:
        if self.is_error:
            return self._v
        return None

    @staticmethod
    def none() -> 'Result':
        return Result()

    def __repr__(self):
        if self.is_error:
            return "<Result({}): {}>".format(type(self._v).__name__, self._v)
        return "<Result: {}>".format(self._v)


__all__ = ["Result"]



from typing import *
import time
import json
import random
import string
from aiohttp.web import WebSocketResponse
from .result import Result
from .utils import create_base58_key


class WebsocketSession:
    def __init__(self):
        self.session_id = create_base58_key(''.join(random.choices(string.ascii_letters, k=12)), length=12, prefix="SS")
        self.create_timestamp = int(time.time() * 1000)
        self.current_user = None
        self.socket: WebSocketResponse = None

    async def prepare(self):
        pass

    async def on_finish(self):
        pass

    async def request(self, message):
        pass

    async def send(self, message):
        if self.socket is not None:
            message_result = self.message_dumps(message)
            if message_result.is_some:
                try:
                    data = message_result.unwrap()
                    if isinstance(data, str):
                        await self.socket.send_str(data)
                    else:
                        await self.socket.send_bytes(data)
                    return True
                except Exception as e:
                    return False
        return False

    async def close(self):
        if self.socket is not None:
            try:
                await self.socket.close()
            except Exception as e:
                pass

    @classmethod
    def message_loads(cls, data) -> Result[Any]:
        try:
            return Result(json.loads(data))
        except Exception as e:
            return Result(e)

    @classmethod
    def message_dumps(cls, o) -> Result[Union[str, bytes]]:
        try:
            return Result(json.dumps(o))
        except Exception as e:
            return Result(e)


__all__ = ["WebsocketSession", ]



from typing import *
from .result import Result
from .error import *


class RedisProtocol:
    CR_LF = [13, 10, ]

    def make_command_buffer(self, *args) -> Result[bytes]:
        if not args:
            return Result(RedisProtocolFormatError())
        buffer = bytearray()
        buffer.extend(self.set_count(len(args)).unwrap())
        for arg in args:
            arg_binary_result = self.set_binary(arg)
            if arg_binary_result.is_error:
                return arg_binary_result
            buffer.extend(arg_binary_result.unwrap())
        return Result(buffer)

    @classmethod
    def set_nil(cls):
        buffer = "$-1\r\n".encode("utf8")
        return Result(buffer)

    @classmethod
    def set_integer(cls, n):
        if not isinstance(n, int):
            return Result(RedisProtocolFormatError())
        return Result(f":{n}\r\n".encode("utf8"))

    @classmethod
    def get_integer(cls, line: bytes):
        if isinstance(line, bytes):
            buffer = bytearray(line)
            if len(buffer) > 1 and buffer.startswith(b":"):
                try:
                    count = int(line[1:])
                    return Result(count)
                except Exception as e:
                    return Result(e)
        return Result(RedisProtocolFormatError())

    @classmethod
    def set_ok(cls, reason: str = "OK"):
        if not isinstance(reason, str):
            return Result(RedisProtocolFormatError())
        return Result(f"+{reason}\r\n".encode("utf8"))

    @classmethod
    def set_err(cls, reason: str = "ERR"):
        if not isinstance(reason, str):
            return Result(RedisProtocolFormatError())
        return Result(f"-{reason}\r\n".encode("utf8"))

    @classmethod
    def set_count(cls, n: int) -> Result[bytes]:
        if not isinstance(n, int):
            return Result(RedisProtocolFormatError())
        return Result(f"*{n}\r\n".encode("utf8"))

    @classmethod
    def set_binary(cls, b: Union[bytes, str, int]):
        if isinstance(b, str):
            arg_buffer = b.encode("utf8")
        elif isinstance(b, bytes):
            arg_buffer = b
        elif isinstance(b, int):
            arg_buffer = str(b).encode("utf8")
        else:
            return Result(RedisProtocolFormatError())
        length = len(arg_buffer)
        buffer = bytearray(f"${length}\r\n".encode("utf8"))
        buffer.extend(arg_buffer)
        buffer.extend(cls.CR_LF)
        return Result(buffer)

    @classmethod
    def get_count(cls, line: bytes) -> Result[int]:
        if isinstance(line, bytes):
            buffer = bytearray(line)
            if len(buffer) > 1 and buffer.startswith(b"*"):
                try:
                    count = int(line[1:])
                    return Result(count)
                except Exception as e:
                    return Result(e)
        return Result(RedisProtocolFormatError())

    @classmethod
    def get_binary_size(cls, line: bytes) -> Result[int]:
        if isinstance(line, bytes):
            buffer = bytearray(line)
            if len(buffer) > 1 and buffer.startswith(b"$"):
                try:
                    count = int(line[1:])
                    if count == -1:
                        return Result(None)
                    return Result(count)
                except Exception as e:
                    return Result(e)
        return Result(RedisProtocolFormatError())


__all__ = ["RedisProtocol", ]



from typing import *
from struct import pack
import time
import base58
import hashlib


class HashCipher(object):
    def __init__(self, salt=None, cipher=hashlib.sha3_256):
        self._salt = salt
        self._cipher = cipher

    def hash(self, text) -> str:
        binary = (self._salt + text).encode("utf8") if self._salt else text.encode("utf8")
        return self._cipher(binary).hexdigest()

    def digest(self, text) -> bytes:
        binary = (self._salt + text).encode("utf8") if self._salt else text.encode("utf8")
        return self._cipher(binary).digest()


def create_base58_key(
        data,
        length: int = 16,
        prefix: str = "",
        salt: Optional[str]= None,
        cipher=hashlib.sha3_256,
        timestamp: bool = False,
) -> str:
    hash_cipher = HashCipher(salt=salt, cipher=cipher)
    hash_key = hash_cipher.digest(str(data))
    slice_key = base58.b58encode(hash_key)[:length]
    if type(slice_key) is bytes:
        slice_key = slice_key.decode("utf8")
    time_part = ''
    if timestamp:
        time_part = base58.b58encode(pack('>Q', int(time.time() * 1000))).zfill(10).decode("utf8")
    key = "{}{}{}".format(prefix, time_part, slice_key)
    return key


__all__ = ["create_base58_key", ]



from setuptools import setup, find_packages

setup(
    name='porkpepper',
    version='0.1.0',
    author='宋伟(songwei)',
    author_email='songwei@songwei.io',
    long_description='',
    url='https://github.com/xdusongwei/porkpepper',
    packages=find_packages(),
    ext_modules=[],
    classifiers=[
        'Programming Language :: Python :: 3',
    ],
    install_requires=[
        'base58',
        'aiohttp',
        'typing',
        'pytest',
        'aioredis',
        'pytest-asyncio',
    ],
    tests_require=[
        'pytest',
    ],
)



from typing import *
import sys
import asyncio
from aiohttp import web
from .websocket_app import WebsocketApp


class PorkPepperNode:
    def __init__(self, redis_server=None, session_class=None, **kwargs):
        self._http_options = kwargs or dict()
        self._redis_options = dict()
        self._http_app: WebsocketApp = None
        self._redis_server = redis_server
        self._session_class = session_class
        self.runner = None

    async def on_start(self):
        pass

    async def on_shutdown(self):
        pass

    async def serve(self, enable_websocket=False, redis_host="127.0.0.1", redis_port=6379, **kwargs):
        redis_server = self._redis_server
        session_class = self._session_class
        app = WebsocketApp(session_class=session_class, **self._http_options)
        app.add_routes([web.route('GET', path, app.handler) for path in ["/porkpepper"]])
        self._http_app = app
        redis_loop = redis_server.serve(redis_host, redis_port)
        redis_task = asyncio.ensure_future(redis_loop)
        runner = web.AppRunner(app)
        await runner.setup()
        await self.on_start()
        try:
            if enable_websocket:
                site = web.TCPSite(runner, **kwargs)
                await site.start()
            while True:
                await asyncio.sleep(60)
        finally:
            redis_task.cancel()
            await runner.cleanup()
            await self.on_shutdown()
